import datetime
import json
import time
import uuid

import redis

__all__ = ['BaseTask']


class BaseTask:
    """
    A base class for defining tasks.
    """

    def __init__(self):

        # Set a unique Id for the task
        self._id = f'{self.get_id_prefix()}:{uuid.uuid4()}'

        # Timestamp the task
        self._timestamp = time.time_ns()

        # The Id of the worker the task is assigned to
        self._assigned_to = None

    def __str__(self):
        return self._id

    @property
    def assigned_to(self):
        return self._assigned_to

    @property
    def id(self):
        return self._id

    @property
    def lock_key(self):
        return f'lock:{self._id}'

    @property
    def reclaim_key(self):
        return f'lock:{self._id}'

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def timestamp_dt(self):
        return datetime.datetime.utcfromtimestamp(self._timestamp / (10 ** 9))

    def assign_to(self, worker_id):
        """Assign the task to a worker"""
        self._assigned_to = worker_id

    def dumps(self):
        """Return a JSON string for the task"""
        return json.dumps(self.to_json_type())

    def to_json_type(self):
        """Return a dictionary containing a JSON safe version of the task"""
        return {
            'id': self._id,
            'assigned_to': self._assigned_to,
            'timestamp': self._timestamp
        }

    def unassign(self):
        """Assign the task to a worker"""
        self._assigned_to = None

    @classmethod
    def from_json_type(cls, data):
        """Convert a dictionary of JSON safe data to a task"""

        assigned_to = data.pop('assigned_to')
        id = data.pop('id')
        timestamp = data.pop('timestamp')

        task = cls(**data)

        task._assigned_to = assigned_to
        task._id = id
        task._timestamp = timestamp

        return task

    @classmethod
    def get_id_prefix(cls):
        """Return a prefix applied to task Ids"""
        return 'swm_task'

    @classmethod
    def loads(cls, s):
        """Load a JSON string as a task"""
        return cls.from_json_type(json.loads(s))



import json
import subprocess

import botocore.exceptions
import botocore.session

from . import BaseSpawner


class LambdaSpawner(BaseSpawner):
    """
    Spawn a new worker on the local system.
    """

    def __init__(self, function_name, payload, client_kwargs=None):

        # The function name to invoke
        self._function_name = function_name

        # The payload to send when invoking the function
        self._payload = payload

        # The keyword arguments to when initializing the lambda client
        self._client_kwargs = client_kwargs or {}


    def spawn(self, amount):

        client = botocore.session.get_session().create_client(
            'lambda',
            **client_kwargs
        )

        for w in range(0, amount):
            client.invoke(
                FunctionName=self._function_name,
                InvocationType='Event',
                Payload=json.dumps(self._payload)
            )



import json
import math
import signal
import time
import traceback
import uuid

import redis

from .events import TaskCompleteEvent, TaskErrorEvent
from .population.controls.static import StaticControl
from .population.spawners.static import StaticSpawner
from .tasks import BaseTask

__all__ = ['BaseWorker']


class BaseWorker:
    """
    A base class for defining worker behaviour.
    """

    def __init__(
        self,
        conn,
        task_cls,
        broadcast_channel='swm_events',
        max_spawn_time=0,
        max_status_interval=600,
        sleep_interval=1,
        idle_lifespan=0,
        population_control=None,
        population_spawner=None,
        population_lock_cool_off=60
    ):

        # The redis connection (instance) to use for communicating with other
        # workers, servers and the monitor/manager.
        self._conn = conn

        # The class of tasks the worker handles
        self._task_cls = task_cls

        # The channel that task events (e.g complete, error) will be broadcast
        # over.
        self._broadcast_channel = broadcast_channel

        # The maximum amount of time allocated to a worker to spawn new
        # workers. If set to 0 the worker will wait forever.
        self._max_spawn_time = max_spawn_time

        # The maximum number of seconds between status updates before the
        # worker is considered to have died.
        self._max_status_interval = max_status_interval

        # The interval between an idle worker checking between tasks
        self._sleep_interval = sleep_interval

        # The workers unique Id (acquired when started)
        self._id = None

        # The maximum period of time a worker can be idle for before it
        # attempts to shut itself down. If 0 the worker will never attempt to
        # shut itself down.
        self._idle_lifespan = idle_lifespan

        # The time the worker was last active (was started or completed a
        # task).
        self._idle_since = None

        # The population control used to regulate the size of the worker
        # population.
        self._population_control = population_control or StaticControl()

        # The population spawner used to spawn new workers
        self._population_spawner = population_spawner or StaticSpawner()

        # The period of wait before the population control lock is released if
        # an error occurred while spawning new workers.
        self._population_lock_cool_off = population_lock_cool_off

    def __str__(self):
        return self._id

    @property
    def id(self):
        return self._id

    def do_task(self, task):
        """Do the given task"""

    def get_tasks(self):
        """Return a list of all relevant tasks for the worker"""

        task_cls_list = self._task_cls
        if isinstance(task_cls_list, BaseTask):
            task_cls_list = [task_cls_list]

        tasks_map = {}

        for task_cls in task_cls_list:

            task_prefix = task_cls.get_id_prefix()
            task_ids = set([
                id for id in self._conn.scan_iter(f'{task_prefix}:*')
            ])

            if not task_ids:
                continue

            tasks = [task_cls.loads(t) for t in self._conn.mget(task_ids)]
            tasks_map.update(zip(task_ids, tasks))

        return tasks_map

    def get_workers(self):
        """Return a list of registered workers"""
        return set([
            id for id in self._conn.scan_iter(f'{self.get_id_prefix()}:*')
        ])

    def start(self):
        """Start the worker"""

        # Set a unique Id for the worker
        self._id = f'{self.get_id_prefix()}:{uuid.uuid4()}'

        # Register the worker
        self._conn.setex(self._id, self._max_status_interval, 'idle')

        # Start the main loop
        try:
            self._idle_since = time.time()
            self._loop()

        except KeyboardInterrupt:
            pass

        finally:
            self.shut_down()

    def shut_down(self):
        """Shutdown the worker"""

        # Unregister the worker
        self._conn.delete(self._id)

    # Event broadcasting and error handling

    def on_complete(self, task_id, data):
        """Broadcast a task completed event"""
        event = TaskCompleteEvent(task_id, data)
        self._conn.publish(self._broadcast_channel, event.dumps())

    def on_error(self, task_id, error):
        """Broadcast a task error event"""
        event = TaskErrorEvent(task_id, str(error))
        self._conn.publish(self._broadcast_channel, event.dumps())

    def on_spawn_error(self, error):
        """Override this method to manage spawn errors"""
        print(traceback.format_exc())
        print(error)

    # Private

    def _loop(self):
        """The application loop"""

        while True:

            # Check to see if the shutdown flag is present for the worker
            # class.
            if self._conn.get(self.get_shutdown_key()):
                return self.shut_down()

            # Get a lists of workers and the tasks to complete
            workers = self.get_workers()
            tasks = self.get_tasks()

            # If the worker is no longer registered then the main loop should
            # be exited.
            if self._id not in workers:
                return

            # Population check (growth and culling)
            population_change = self._population_control.population_change(
                workers,
                tasks
            )
            population_lock_key = self.get_population_lock_key()
            time_idle = time.time() - self._idle_since

            self._conn.delete(population_lock_key)

            if population_change > 0:

                # Attempt to get a population lock so we can spawn new workers
                if self._conn.setnx(population_lock_key, self.id):
                    try:
                        self._population_spawner.spawn(population_change)

                        # Wait until the new workers have spawned
                        spawned_at = time.time()
                        while True:
                            added_workers = self.get_workers() - workers

                            if len(added_workers) >= population_change:
                                break

                            spawning = time.time() - spawned_at
                            if spawning > (self._max_spawn_time or math.inf):
                                break

                            time.sleep(1)

                        self._conn.delete(population_lock_key)

                    except Exception as e:

                        # Report the error but allow the worker to continue to
                        # run.
                        self.on_spawn_error(error)

                        # Delay removing the population lock to ensure we
                        # don't end up in a race condition between
                        self._conn.expire(
                            population_lock_key,
                            self._population_lock_cool_off
                        )
                        time.sleep(self._population_lock_cool_off)

            elif population_change < 0 \
                    and time_idle > (self._idle_lifespan or math.inf):

                # Attempt to get a population lock so we can shut this worker
                # down.
                if self._conn.setnx(population_lock_key, self.id):
                    try:
                        return self.shut_down()

                    finally:
                        self._conn.delete(population_lock_key)

            # Find any tasks that are currently assigned to workers that are
            # no longer registered and convert them to pending tasks.
            for task in tasks.values():

                # Check for any task that is assigned to a non-existent worker
                if task.assigned_to and task.assigned_to not in workers:

                    if self._conn.setnx(task.reclaim_key, self._id):

                        # Clear the existing lock for the task
                        self._conn.expire(task.reclaim_key, 10)
                        self._conn.delete(task.lock_key)

                        # Unassign the task so it will be handled by this
                        # worker.
                        task.unassign()

                # Check if the task is pending
                if not task.assigned_to:

                    if self._conn.setnx(task.lock_key, self.id):

                        # Assign the task to this worker
                        task.assign_to(self._id)
                        self._conn.set(task._id, task.dumps())

                        # Update the workers status to busy
                        self._conn.setex(
                            self._id,
                            self._max_status_interval,
                            'busy'
                        )

                        # Process the task
                        try:
                            event_data = self.do_task(task)
                            self.on_complete(task.id, event_data)

                        except Exception as e:
                            self.on_error(task.id, e)

                        finally:

                            # Delete the task and associated lock
                            self._conn.delete(task.id)
                            self._conn.delete(task.lock_key)

                            # Record the time at which the worker became idle
                            self._idle_since = time.time()

            # Update the workers status to idle
            self._conn.setex(self._id, self._max_status_interval, 'idle')

            time.sleep(self._sleep_interval)

    # Class methods

    @classmethod
    def get_id_prefix(cls):
        """Return a prefix applied to worker Ids"""
        return 'swm_worker'

    @classmethod
    def get_population_lock_key(cls):
        """
        Return the lock key for this class or workers to use when managing the
        population.
        """
        return f'{cls.get_id_prefix()}_population_contol'

    @classmethod
    def get_shutdown_key(cls):
        """Return the key used to flag to workers that they should shutdown"""
        return f'{cls.get_id_prefix()}_shutdown'



import os
import subprocess

from . import BaseSpawner


class LocalSpawner(BaseSpawner):
    """
    Spawn a new worker on the local system.
    """

    def __init__(self, args):

        # The args to use to spawn a new worker
        self._args = args

    def spawn(self, amount):
        for w in range(0, amount):
            subprocess.Popen(self._args, preexec_fn=os.setpgrp)



import json

__all__ = [
    'BaseEvent',
    'TaskCompleteEvent',
    'TaskErrorEvent'
]


class BaseEvent:
    """
    A base event class.
    """

    def __init__(self, type):

        # The type of event
        self._type = type

    def __str__(self):
        return self._type

    @property
    def type(self):
        return self._type

    def dumps(self):
        """Return a JSON string for the event"""
        return json.dumps(self.to_json_type())

    def to_json_type(self):
        """Return a dictionary containing a JSON safe version of the event"""
        return {'type': self._type}

    @classmethod
    def from_json_type(cls, data):
        """Convert a dictionary of JSON safe data to an event"""
        return cls(**data)

    @classmethod
    def loads(cls):
        """Load a JSON string as an event"""
        return cls.from_json_type(json.loads(s))


class TaskCompleteEvent(BaseEvent):
    """
    An event published when a task is completed.
    """

    def __init__(self, task_id, data):
        super().__init__('task_complete')

        # The Id of task that was completed
        self._task_id = task_id

        # The data associated with the event (typically this forms the body
        # of the response to the caller).
        self._data = data

    @property
    def data(self):
        return self._data

    @property
    def task_id(self):
        return self._task_id

    def to_json_type(self):
        """Return a dictionary containing a JSON safe version of the event"""
        return {
            'data': self._data,
            'task_id': self.task_id,
            'type': self._type
        }

    @classmethod
    def from_json_type(cls, data):
        """Convert a dictionary of JSON safe data to an event"""
        data.pop('type')
        return cls(**data)


class TaskErrorEvent(BaseEvent):
    """
    An event published when an error occurs attempting to do task.
    """

    def __init__(self, task_id, reason):
        super().__init__('task_error')

        # The Id of task that was completed
        self._task_id = task_id

        # The reason for the error
        self._reason = reason

    @property
    def reason(self):
        return self._reason

    @property
    def task_id(self):
        return self._task_id

    def to_json_type(self):
        """Return a dictionary containing a JSON safe version of the event"""
        return {
            'reason': self._reason,
            'task_id': self._task_id,
            'type': self._type
        }

    @classmethod
    def from_json_type(cls, data):
        """Convert a dictionary of JSON safe data to an event"""
        data.pop('type')
        return cls(**data)



"""
Utils for monitoring workers and tasks.
"""

import operator
import time

__all__ = [
    'get_tasks',
    'get_workers',
    'shutdown_workers'
]


def get_tasks(conn, *task_cls_list):
    """Return a list of incomplete tasks in the heap"""

    tasks = []
    for task_cls in task_cls_list:

        task_prefix = task_cls.get_id_prefix()
        task_ids = set([id for id in conn.scan_iter(f'{task_prefix}:*')])

        if not task_ids:
            continue

        tasks.extend([task_cls.loads(t) for t in conn.mget(task_ids)])

    return sorted(tasks, key=operator.attrgetter('timestamp'))

def get_workers(conn, worker_cls):
    """Return a list of active workers [(worker_id, status)]"""
    worker_prefix = worker_cls.get_id_prefix()
    worker_ids = set([id for id in conn.scan_iter(f'{worker_prefix}:*')])
    return sorted(
        zip(worker_ids, conn.mget(worker_ids)),
        key=operator.itemgetter(0)
    )

def shutdown_workers(conn, worker_cls):
    """Shutdown all workers for the given worker class"""

    try:
        conn.set(worker_cls.get_shutdown_key(), 'shutdown')

        while True:
            workers = list(conn.scan_iter(f'{worker_cls.get_id_prefix()}:*'))
            if len(workers) == 0:
                break

            time.sleep(1)

    finally:
        conn.delete(worker_cls.get_shutdown_key())



import asyncio
import json

import aioredis
import tornado.locks
import tornado.web

from swm.events import TaskCompleteEvent, TaskErrorEvent

__all__ = [
    'Application',
    'RequestHandler',
    'TaskEventListener'
]


class Application(tornado.web.Application):

    @property
    def redis(self):
        return self._redis

    @property
    def redis_sub(self):
        return self._redis_sub

    @property
    def task_event_listener(self):
        return self._task_event_listener

    async def listen_for_task_events(self, conn, channel='swm_events'):
        """"""
        self._task_event_listener = TaskEventListener()
        await self._task_event_listener.listen(conn, channel)


class RequestHandler(tornado.web.RequestHandler):

    def initialize(self):
        self.wait_future = None

    @property
    def redis(self):
        return self.application.redis

    @property
    def task_event_listener(self):
        return self.application.task_event_listener

    async def add_task_and_wait(self, task):
        """Add a task to the heap and wait for an event"""
        await self.redis.set(task.id, json.dumps(task.to_json_type()))

        event = None
        while not event:
            self.wait_future = self.task_event_listener.wait()
            
            try:
                await self.wait_future
            except asyncio.CancelledError:
                return
            
            event = self.task_event_listener.get_event(task.id) 

        if self.request.connection.stream.closed():
            return

        return event

    async def add_task_and_forget(self, task):
        """Add a task to the heap and return (e.g don't wait)"""
        await self.redis.set(task.id, json.dumps(task.to_json_type()))

    def on_connection_close(self):
        # Cancel any wait future
        if self.wait_future:
            self.wait_future.cancel()


class TaskEventListener:
    """
    The `TaskEventListener` class provides a fire and wait mechanism for 
    request handlers posting tasks to the worker network. 
    """

    EVENT_TYPES = {
        'task_complete': TaskCompleteEvent,
        'task_error': TaskErrorEvent
    }

    def __init__(self):

        # The lock used to allow request handlers to listen for a task event 
        # from the worker network.
        self._condition = tornado.locks.Condition()

        # The last event received by the listener
        self._last_event = None

    def get_event(self, task_id):
        """Return the last event if it matches the given task Id"""
        if self._last_event:
            if self._last_event.task_id == task_id:
                return self._last_event

    async def listen(self, conn, channel='swm_events'):
        """Listen for task events"""
        asyncio.ensure_future(
            self._receive((await conn.subscribe(channel))[0])
        )

    def wait(self):
        """Wait for a task event"""
        return self._condition.wait()

    async def _receive(self, channel):
        """Handle receiving an event (message) from redis"""

        while await channel.wait_message():
            data = await channel.get_json()

            # Check we received a know event type
            event_cls = self.EVENT_TYPES.get(data.get('type'))
            if event_cls:

                # Store the event
                self._last_event = event_cls.from_json_type(data)

                # Notify all listeners that a new event has been received
                self._condition.notify_all()
